@@ -1,13 +1,14 @@
 import sys
 import struct
-import binascii
 
 if sys.version < '3':
-    from btcommon import *
+    from .btcommon import *
     import _bluetooth as _bt
+    get_byte = ord
 else:
     from bluetooth.btcommon import *
     import bluetooth._bluetooth as _bt
+    get_byte = int
 import array
 import fcntl
 _constants = [ 'HCI', 'RFCOMM', 'L2CAP', 'SCO', 'SOL_L2CAP', 'SOL_RFCOMM',\
@@ -19,10 +20,15 @@ del _constants
 
 # ============== SDP service registration and unregistration ============
 
-def discover_devices (duration=8, flush_cache=True, lookup_names=False, lookup_class=False):
-    sock = _gethcisock ()
+def discover_devices (duration=8, flush_cache=True, lookup_names=False,
+                      lookup_class=False, device_id=-1):
+    if device_id == -1:
+        device_id = _bt.hci_get_route()
+
+    sock = _gethcisock (device_id)
     try:
-        results = _bt.hci_inquiry (sock, duration=duration, flush_cache=True, lookup_class=lookup_class)
+        results = _bt.hci_inquiry (sock, duration=duration, flush_cache=True,
+                                   lookup_class=lookup_class, device_id=device_id)
     except _bt.error:
         sock.close ()
         raise BluetoothError ("error communicating with local "
@@ -32,13 +38,13 @@ def discover_devices (duration=8, flush_cache=True, lookup_names=False, lookup_c
         pairs = []
         for item in results:
             if lookup_class:
-               addr, dev_class = item
+                addr, dev_class = item
             else:
-               addr = item
+                addr = item
             timeoutms = int (10 * 1000)
             try: 
                 name = _bt.hci_read_remote_name (sock, addr, timeoutms)
-            except _bt.error as e:
+            except _bt.error:
                 # name lookup failed.  either a timeout, or I/O error
                 continue
             pairs.append ((addr, name, dev_class) if lookup_class else (addr, name))
@@ -56,7 +62,7 @@ def lookup_name (address, timeout=10):
     timeoutms = int (timeout * 1000)
     try: 
         name = _bt.hci_read_remote_name (sock, address, timeoutms)
-    except _bt.error as e:
+    except _bt.error:
         # name lookup failed.  either a timeout, or I/O error
         name = None
     sock.close ()
@@ -77,7 +83,7 @@ def set_packet_timeout (address, timeout):
     this method
     """
     n = round (timeout / 0.625)
-    _write_flush_timeout (address, n)
+    write_flush_timeout (address, n)
 
 def get_l2cap_options (sock):
     """get_l2cap_options (sock, mtu)
@@ -116,7 +122,7 @@ def set_l2cap_mtu (sock, mtu):
     options[0] = options[1] = mtu
     set_l2cap_options (sock, options)
 
-def _get_available_port (protocol):
+def _get_available_port (protocol, addr=""):
     """
     deprecated.  bind to PORT_ANY instead.
     """
@@ -124,7 +130,7 @@ def _get_available_port (protocol):
         for channel in range (1,31):
             s = BluetoothSocket (RFCOMM)
             try:
-                s.bind ( ("", channel))
+                s.bind ( (addr, channel))
                 s.close ()
                 return channel
             except:
@@ -133,7 +139,7 @@ def _get_available_port (protocol):
         for psm in range (0x1001,0x8000,2):
             s = BluetoothSocket (L2CAP)
             try:
-                s.bind ( ("", psm))
+                s.bind ( (addr, psm))
                 s.close ()
                 return psm
             except:
@@ -168,7 +174,7 @@ class BluetoothSocket:
     def bind (self, addrport):
         if self._proto == RFCOMM or self._proto == L2CAP:
             addr, port = addrport
-            if port == 0: addrport = (addr, _get_available_port (self._proto))
+            if port == 0: addrport = (addr, _get_available_port (self._proto, addr))
         return self._sock.bind (addrport)
 
     def get_l2cap_options(self):
@@ -266,8 +272,7 @@ def find_service (name = None, uuid = None, address = None):
                 continue
 
             if name is not None:
-                matches = filter (lambda s: s.get ("name", "") == name, \
-                        matches)
+                matches = [s for s in matches if s.get ("name", "") == name]
 
             for m in matches:
                 m["host"] = addr
@@ -286,42 +291,42 @@ def _gethcisock (device_id = -1):
         raise BluetoothError ("error accessing bluetooth device")
     return sock
 
-def _get_acl_conn_handle (hci_sock, addr):
+def get_acl_conn_handle (hci_sock, addr):
     hci_fd = hci_sock.fileno ()
     reqstr = struct.pack ("6sB17s", _bt.str2ba (addr), 
-            _bt.ACL_LINK, "\0" * 17)
-    request = array.array ("c", reqstr)
+            _bt.ACL_LINK, b"\0" * 17)
+    request = array.array ("b", reqstr)
     try:
         fcntl.ioctl (hci_fd, _bt.HCIGETCONNINFO, request, 1)
-    except IOError as e:
+    except IOError:
         raise BluetoothError ("There is no ACL connection to %s" % addr)
 
     # XXX should this be "<8xH14x"?
     handle = struct.unpack ("8xH14x", request.tostring ())[0]
     return handle
 
-def _write_flush_timeout (addr, timeout):
+def write_flush_timeout (addr, timeout):
     hci_sock = _bt.hci_open_dev ()
     # get the ACL connection handle to the remote device
-    handle = _get_acl_conn_handle (hci_sock, addr)
+    handle = get_acl_conn_handle (hci_sock, addr)
     # XXX should this be "<HH"
     pkt = struct.pack ("HH", handle, _bt.htobs (timeout))
     response = _bt.hci_send_req (hci_sock, _bt.OGF_HOST_CTL, 
         0x0028, _bt.EVT_CMD_COMPLETE, 3, pkt)
-    status = struct.unpack ("B", response[0])[0]
+    status = get_byte(response[0])
     rhandle = struct.unpack ("H", response[1:3])[0]
     assert rhandle == handle 
     assert status == 0
 
-def _read_flush_timeout (addr):
+def read_flush_timeout (addr):
     hci_sock = _bt.hci_open_dev ()
     # get the ACL connection handle to the remote device
-    handle = _get_acl_conn_handle (hci_sock, addr)
+    handle = get_acl_conn_handle (hci_sock, addr)
     # XXX should this be "<H"?
     pkt = struct.pack ("H", handle)
     response = _bt.hci_send_req (hci_sock, _bt.OGF_HOST_CTL, 
         0x0027, _bt.EVT_CMD_COMPLETE, 5, pkt)
-    status = struct.unpack ("B", response[0])[0]
+    status = get_byte(response[0])
     rhandle = struct.unpack ("H", response[1:3])[0]
     assert rhandle == handle
     assert status == 0
@@ -329,6 +334,12 @@ def _read_flush_timeout (addr):
     return fto
 
 # =============== DeviceDiscoverer ==================
+def byte_to_signed_int(byte_):
+    if byte_ > 127:
+        return byte_ - 256
+    else:
+        return byte_
+
 class DeviceDiscoverer:
     """
     Skeleton class for finer control of the device discovery process.
@@ -423,9 +434,9 @@ class DeviceDiscoverer:
             try:
                 _bt.hci_send_cmd (self.sock, _bt.OGF_LINK_CTL, \
                         _bt.OCF_INQUIRY_CANCEL)
+            except:
                 self.sock.close ()
                 self.sock = None
-            except:
                 raise BluetoothError ("error canceling inquiry")
             self.is_inquiring = False
 
@@ -445,15 +456,13 @@ class DeviceDiscoverer:
         self._process_hci_event ()
 
     def _process_hci_event (self):
-        import socket
-
         if self.sock is None: return
         # voodoo magic!!!
         pkt = self.sock.recv (258)
         ptype, event, plen = struct.unpack ("BBB", pkt[:3])
         pkt = pkt[3:]
         if event == _bt.EVT_INQUIRY_RESULT:
-            nrsp = struct.unpack ("B", pkt[0])[0]
+            nrsp = get_byte(pkt[0])
             for i in range (nrsp):
                 addr = _bt.ba2str (pkt[1+6*i:1+6*i+6])
                 psrm = pkt[ 1+6*nrsp+i ]
@@ -466,9 +475,9 @@ class DeviceDiscoverer:
                 clockoff = pkt[1+12*nrsp+2*i:1+12*nrsp+2*i+2]
 
                 self._device_discovered (addr, devclass, 
-                        psrm, pspm, clockoff, None)
+                        psrm, pspm, clockoff, None, None)
         elif event == _bt.EVT_INQUIRY_RESULT_WITH_RSSI:
-            nrsp = struct.unpack ("B", pkt[0])[0]
+            nrsp = get_byte(pkt[0])
             for i in range (nrsp):
                 addr = _bt.ba2str (pkt[1+6*i:1+6*i+6])
                 psrm = pkt[ 1+6*nrsp+i ]
@@ -481,12 +490,12 @@ class DeviceDiscoverer:
                         (devclass_raw[1] << 8) | \
                         devclass_raw[0]
                 clockoff = pkt[1+11*nrsp+2*i:1+11*nrsp+2*i+2]
-                rssi = struct.unpack ("b", pkt[1+13*nrsp+i])[0]
+                rssi = byte_to_signed_int(get_byte(pkt[1+13*nrsp+i]))
 
                 self._device_discovered (addr, devclass, 
-                        psrm, pspm, clockoff, None)
+                        psrm, pspm, clockoff, rssi, None)
         elif _bt.HAVE_EVT_EXTENDED_INQUIRY_RESULT and event == _bt.EVT_EXTENDED_INQUIRY_RESULT:
-            nrsp = struct.unpack ("B", pkt[0])[0]
+            nrsp = get_byte(pkt[0])
             for i in range (nrsp):
                 addr = _bt.ba2str (pkt[1+6*i:1+6*i+6])
                 psrm = pkt[ 1+6*nrsp+i ]
@@ -497,29 +506,29 @@ class DeviceDiscoverer:
                         (devclass_raw[1] << 8) | \
                         devclass_raw[0]
                 clockoff = pkt[1+11*nrsp+2*i:1+11*nrsp+2*i+2]
-                rssi = struct.unpack ("b", pkt[1+13*nrsp+i])[0]
+                rssi = byte_to_signed_int(get_byte(pkt[1+13*nrsp+i]))
 
                 data_len = _bt.EXTENDED_INQUIRY_INFO_SIZE - _bt.INQUIRY_INFO_WITH_RSSI_SIZE
                 data = pkt[1+14*nrsp+i:1+14*nrsp+i+data_len]
                 name = None
                 pos = 0
                 while(pos <= len(data)):
-                    struct_len = ord(data[pos])
+                    struct_len = get_byte(data[pos])
                     if struct_len == 0:
                         break
-                    eir_type = ord(data[pos+1])
+                    eir_type = get_byte(data[pos+1])
                     if eir_type == 0x09: # Complete local name
                         name = data[pos+2:pos+struct_len+1]
                     pos += struct_len + 2
 
                 self._device_discovered (addr, devclass,
-                        psrm, pspm, clockoff, name)
-        elif event == _bt.EVT_INQUIRY_COMPLETE:
+                        psrm, pspm, clockoff, rssi, name)
+        elif event == _bt.EVT_INQUIRY_COMPLETE or event == _bt.EVT_CMD_COMPLETE:
             self.is_inquiring = False
             if len (self.names_to_find) == 0:
 #                print "inquiry complete (evt_inquiry_complete)"
                 self.sock.close ()
-                self.inquiry_complete ()
+                self._inquiry_complete ()
             else:
                 self._send_next_name_req ()
 
@@ -533,9 +542,9 @@ class DeviceDiscoverer:
 #                print "inquiry complete (bad status 0x%X 0x%X 0x%X)" % \
 #                        (status, ncmd, opcode)
                 self.names_to_find = {}
-                self.inquiry_complete ()
+                self._inquiry_complete ()
         elif event == _bt.EVT_REMOTE_NAME_REQ_COMPLETE:
-            status = struct.unpack ("B", pkt[0])[0]
+            status = get_byte(pkt[0])
             addr = _bt.ba2str (pkt[1:7])
             if status == 0:
                 try:
@@ -543,10 +552,10 @@ class DeviceDiscoverer:
                 except IndexError:
                     name = '' 
                 if addr in self.names_to_find:
-                    device_class = self.names_to_find[addr][0]
-                    self.device_discovered (addr, device_class, name)
+                    device_class, rssi = self.names_to_find[addr][:2]
+                    self.device_discovered (addr, device_class, rssi, name)
                     del self.names_to_find[addr]
-                    self.names_found[addr] = ( device_class, name)
+                    self.names_found[addr] = ( device_class, rssi, name)
                 else:
                     pass
             else:
@@ -566,24 +575,24 @@ class DeviceDiscoverer:
 #            print "unrecognized packet type 0x%02x" % ptype
 
     def _device_discovered (self, address, device_class, 
-            psrm, pspm, clockoff, name):
+            psrm, pspm, clockoff, rssi, name):
         if self.lookup_names:
             if name is not None:
-                self.device_discovered (address, device_class, name)
+                self.device_discovered (address, device_class, rssi, name)
             elif address not in self.names_found and \
                 address not in self.names_to_find:
             
                 self.names_to_find[address] = \
-                    (device_class, psrm, pspm, clockoff)
+                    (device_class, rssi, psrm, pspm, clockoff)
         else:
-            self.device_discovered (address, device_class, None)
+            self.device_discovered (address, device_class, rssi, None)
 
     def _send_next_name_req (self):
         assert len (self.names_to_find) > 0
-        address = self.names_to_find.keys ()[0]
-        device_class, psrm, pspm, clockoff = self.names_to_find[address]
-        bdaddr = _bt.str2ba (address)
-        
+        address = list(self.names_to_find.keys ())[0]
+        device_class, rssi, psrm, pspm, clockoff = self.names_to_find[address]
+        bdaddr = _bt.str2ba (address) #TODO not supported in python3
+
         cmd_pkt = "%s%s\0%s" % (bdaddr, psrm, clockoff)
 
         try:
@@ -605,7 +614,7 @@ class DeviceDiscoverer:
         This method exists to be overriden
         """
 
-    def device_discovered (self, address, device_class, name):
+    def device_discovered (self, address, device_class, rssi, name):
         """
         Called when a bluetooth device is discovered.
 
@@ -622,10 +631,20 @@ class DeviceDiscoverer:
         [1] https://www.bluetooth.org/foundry/assignnumb/document/baseband
         """
         if name:
-            print("found: %s - %s (class 0x%X)" % \
-                    (address, name, device_class))
+            print(("found: %s - %s (class 0x%X, rssi %s)" % \
+                    (address, name, device_class, rssi)))
         else:
-            print("found: %s (class 0x%X)" % (address, device_class))
+            print(("found: %s (class 0x%X)" % (address, device_class)))
+            print(("found: %s (class 0x%X, rssi %s)" % \
+                    (address, device_class, rssi)))
+
+    def _inquiry_complete (self):
+        """
+        Called when an inquiry started by find_devices has completed.
+        """
+        self.sock.close ()
+        self.sock = None
+        self.inquiry_complete()
 
     def inquiry_complete (self):
         """
